﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Reflection;
using System.Windows.Forms; // for dll browsing dialog
using System.Runtime.Serialization.Json;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;

// Asprise C# VB.NET Imaging and Scanning SDK Library API. <a href='http://asprise.com/c%23-vb.net-scanner-api/twain-wia-document-scan-library-sdk-samples-docs.html' target='_blank'>Read the developer's guide here.</a>
namespace asprise_imaging_api
{
    /// <summary>
    /// Represents an Asprise OCR engine.
    /// <a href='http://asprise.com/c%23-vb.net-scanner-api/twain-wia-document-scan-library-sdk-samples-docs.html' target='_blank'>Read the developer's guide here.</a>
    /// </summary>
    public class AspriseImaging
    {
        private const string DLL_NAME_32 = "asprise_imaging.dll";
        private const string DLL_NAME_64 = "asprise_imaging_x64.dll";

        /// <summary>Turn off logging </summary>
        public const int LOG_LEVEL_OFF = 0;
        /// <summary>Log level: Error </summary>
        public const int LOG_LEVEL_ERROR = 1;
        /// <summary>Log level: Warn </summary>
        public const int LOG_LEVEL_WARN = 2;
        /// <summary>Log level: Info (Default) </summary>
        public const int LOG_LEVEL_INFO = 3;
        /// <summary>Log level: Debug </summary>
        public const int LOG_LEVEL_DEBUG = 4;

        /// <summary> Log to standard output </summary>
        public const String LOG_TO_STDOUT = "stdout";
        /// <summary> Log to standard error </summary>
        public const String LOG_TO_STDERR = "stderr";
        
        /// <summary>Return original image in base64</summary>
        public const String OUTPUT_RETURN_BASE64 = "return-base64";
        /// <summary>Return image thumbnail in base64</summary>
        public const String OUTPUT_RETURN_BASE64_THUMB = "return-base64-thumbnail";
        /// <summary>Return handle to original image</summary>
        public const String OUTPUT_RETURN_HANDLE = "return-handle";
        /// <summary>Return handle to image thumbnail</summary>
        public const String OUTPUT_RETURN_HANDLE_THUMB = "return-handle-thumbnail";
        /// <summary>Save original image to local disk</summary>
        public const String OUTPUT_SAVE = "save";
        /// <summary>Save image thumbnail to local disk</summary>
        public const String OUTPUT_SAVE_THUMB = "save-thumbnail";
        /// <summary>Upload original image</summary>
        public const String OUTPUT_UPLOAD = "upload";
        /// <summary>Upload image thumbnail</summary>
        public const String OUTPUT_UPLOAD_THUMB = "upload-thumbnail";

        /// <summary>JPEG image format</summary>
        public const String FORMAT_JPG = "jpg";
        /// <summary>PNG image format</summary>
        public const String FORMAT_PNG = "png";
        /// <summary>Bitmap image format</summary>
        public const String FORMAT_BMP = "bmp";
        /// <summary>TIFF image format</summary>
        public const String FORMAT_TIF = "tif";
        /// <summary>PDF</summary>
        public const String FORMAT_PDF = "pdf";

        /// <summary>TIFF compression: G3</summary>
        public const String TIFF_COMPRESSION_CCITT_G3 = "G3";
        /// <summary>TIFF compression: G4</summary>
        public const String TIFF_COMPRESSION_CCITT_G4 = "G4";
        /// <summary>TIFF compression: LZW</summary>
        public const String TIFF_COMPRESSION_LZW	= "LZW";
        /// <summary>TIFF compression: RLE</summary>
        public const String TIFF_COMPRESSION_RLE	= "RLE";
        /// <summary>TIFF compression: NONE</summary>
        public const String TIFF_COMPRESSION_NONE = "NONE";

        /// <summary>TIFF compression: Packbits</summary>
        public const String TIFF_COMPRESSION_PACKBITS = "PACKBITS";
        /// <summary>TIFF compression: Zip</summary>
        public const String TIFF_COMPRESSION_ZIP	= "ZIP";

        /// <summary>EXIF: DocumentName</summary>
        public const String EXIF_NAME_DocumentName = "DocumentName";
        /// <summary>EXIF: ImageDescription</summary>
        public const String EXIF_NAME_ImageDescription = "ImageDescription";
        /// <summary>EXIF: EquipMake</summary>
        public const String EXIF_NAME_EquipMake = "EquipMake";
        /// <summary>EXIF: EquipModel</summary>
        public const String EXIF_NAME_EquipModel	= "EquipModel";
        /// <summary>EXIF: Copyright</summary>
        public const String EXIF_NAME_Copyright = "Copyright";
        /// <summary>EXIF: UserComment</summary>
        public const String EXIF_NAME_UserComment = "UserComment";

        public const String FUNC_twain_list_sources = "asprise_scan_twain_list_sources";
        public const String FUNC_twain_get_default_source_name = "asprise_scan_twain_get_default_source_name";
        public const String FUNC_twain_select_default_source = "asprise_scan_twain_select_default_source";
        public const String FUNC_twain_scan = "asprise_scan_twain_scan";
        public const String FUNC_image_output = "asprise_scan_image_output";
        public const String FUNC_image_info = "asprise_scan_image_info";
        public const String FUNC_image_process = "asprise_scan_image_process";
        public const String FUNC_twain_source_manager_open = "asprise_scan_twain_source_manager_open";
        public const String FUNC_twain_source_manager_close = "asprise_scan_twain_source_manager_close";
        public const String FUNC_twain_open_data_source = "asprise_scan_twain_open_data_source";
        public const String FUNC_twain_close_data_source = "asprise_scan_twain_close_data_source";
        public const String FUNC_twain_get_source = "asprise_scan_twain_get_source";
        public const String FUNC_twain_caps_set = "asprise_scan_twain_caps_set";
        public const String FUNC_twain_down_to_state_gracefully = "asprise_scan_twain_down_to_state_gracefully";
        public const String FUNC_twain_cap_set = "asprise_scan_twain_cap_set";
        public const String FUNC_twain_cap_get = "asprise_scan_twain_cap_get";
        public const String FUNC_twain_cap_reset = "asprise_scan_twain_cap_reset";
        public const String FUNC_twain_caps_print = "asprise_scan_twain_caps_print";
        public const String FUNC_twain_create_dummy_window = "asprise_scan_twain_create_dummy_window";
        public const String FUNC_twain_destroy_window = "asprise_scan_twain_destroy_window";
        public const String FUNC_twain_register_app = "asprise_scan_twain_register_app";
        public const String FUNC_twain_get_state = "asprise_scan_twain_get_state";
        public const String FUNC_twain_is_loaded = "asprise_scan_twain_is_loaded";
        public const String FUNC_version = "asprise_scan_version";
        public const String FUNC_buildInfo = "asprise_scan_build_info";

        static AspriseImaging()
        {
            getDll();
        }

        /// <summary>
        /// Unmanaged code access (32bit).
        /// </summary>
        private static class ImgDll32
        {
            [DllImport(DLL_NAME_32, CharSet = CharSet.Ansi)]
            public static extern IntPtr asprise_scan_version();

            [DllImport(DLL_NAME_32, CharSet = CharSet.Ansi)]
            public static extern IntPtr asprise_scan_build_info();

            [DllImport(DLL_NAME_32, CharSet = CharSet.Ansi)]
            public static extern IntPtr asprise_scan_call_func(string json);

            [DllImport(DLL_NAME_32, CharSet = CharSet.Ansi)]
            public static extern void asprise_scan_input_license(string licenseeName, string licenseCode);

            [DllImport(DLL_NAME_32, CharSet = CharSet.Ansi)]
            public static extern void asprise_scan_util_delete_chars(Int64 handle);

            [DllImport(DLL_NAME_32, CharSet = CharSet.Ansi)]
            public static extern void asprise_scan_configure_logging(int level, string outputFile);
        }

        /// <summary>
        /// Unmanaged code access (64bit).
        /// </summary>
        private static class ImgDll64
        {
            [DllImport(DLL_NAME_64, CharSet = CharSet.Ansi)]
            public static extern IntPtr asprise_scan_version();

            [DllImport(DLL_NAME_64, CharSet = CharSet.Ansi)]
            public static extern IntPtr asprise_scan_build_info();

            [DllImport(DLL_NAME_64, CharSet = CharSet.Ansi)]
            public static extern IntPtr asprise_scan_call_func(string json);

            [DllImport(DLL_NAME_64, CharSet = CharSet.Ansi)]
            public static extern void asprise_scan_input_license(string licenseeName, string licenseCode);

            [DllImport(DLL_NAME_64, CharSet = CharSet.Ansi)]
            public static extern void asprise_scan_util_delete_chars(Int64 handle);

            [DllImport(DLL_NAME_64, CharSet = CharSet.Ansi)]
            public static extern void asprise_scan_configure_logging(int level, string outputFile);
        }

        private Thread threadDoingScan;

        private String appId;
        private int windowHandle;

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="control">Optional control whose form will be used as the dialog owner.</param>
        public AspriseImaging(Control control) : this("CS", getWindowHandle(control))
        {

        }

        /// <summary>
        /// Constructor.
        /// </summary>
        public AspriseImaging() : this("CS", 0)
        {

        }

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="appId">Optional name of your application</param>
        /// <param name="windowHandle">Handle to optional window to be used as the dialog owner or 0 if no available; </param>
        public AspriseImaging(String appId, int windowHandle)
        {
            this.appId = appId == null ? "CS" : appId;
            this.windowHandle = windowHandle;
        }

        private volatile Thread calledInThread;

        protected void ensureScanFuncsCallInTheSameThread()
        {
            if (calledInThread == null)
            {
                calledInThread = Thread.CurrentThread;
            }
            else
            {
                if (calledInThread != Thread.CurrentThread)
                {
                    throw new SystemException("Scan operations should be performed in the same thread. Previous used thread: " + calledInThread.Name + ", current thread: " + Thread.CurrentThread.Name);
                }
            }
        }

        public static int getWindowHandle(System.Windows.Forms.Control control) {
            Form form = null;
            if (control is Form)
            {
                form = (Form)control;
            }
            else if (control != null)
            {
                form = control.FindForm();
            }

            if (form == null)
            {
                return 0;
            }
            return form.Handle.ToInt32(); // http://stackoverflow.com/questions/10621437/how-to-get-a-hwnd-handle-out-of-a-system-windows-forms-form
        }


    /// <summary>
    /// Performs scanning from a device and output (return, save, and/or upload).
    /// </summary>
    /// <param name="scanRequest">scan request object</param>
    /// <param name="sourceName">the exact source name or "select" to prompt dialog selection; "default" to use default source; "current" refers to current opened source if any.</param>
    /// <param name="showUI">set to true to use scanner UI or false to hide the UI. Set to true for maximum compatibility.</param>
    /// <param name="modalUI">whether the scanner UI should be modal. Set to to true if you are not sure.</param>
    /// <returns>Scan result or null if user cancels.</returns>
    public Result Scan(Request scanRequest, String sourceName, Boolean showUI, Boolean modalUI)
    {
        return this.Scan(scanRequest.ToJson(), sourceName, showUI, modalUI);
    }

    /// <summary>
    /// Performs scanning from a device and output (return, save, and/or upload).
    /// </summary>
    /// <param name="scanRequestInJson">scan request in JSON format.</param>
    /// <param name="sourceName">the exact source name or "select" to prompt dialog selection; "default" to use default source; "current" refers to current opened source if any.</param>
    /// <param name="showUI">set to true to use scanner UI or false to hide the UI. Set to true for maximum compatibility.</param>
    /// <param name="modalUI">whether the scanner UI should be modal. Set to to true if you are not sure.</param>
    /// <returns>Scan result or null if user cancels.</returns>
    public Result Scan(String scanRequestInJson, String sourceName, Boolean showUI, Boolean modalUI)
    {
        ensureScanFuncsCallInTheSameThread();
        String rawResult = ScanAndReturnRaw(scanRequestInJson, sourceName, showUI, modalUI);
        if (rawResult == null)
        {
            return null; // user cancels
        }
        if (rawResult != null && rawResult.StartsWith("<error:"))
        {
            throw new ImagingException(rawResult);
        }
        return new Result(ScanUtils.parseJson(rawResult));
    }
        
    /// <summary>
    /// Performs scanning from a device and output result in JSON.
    /// </summary>
    /// <param name="scanRequestInJson">scan request in JSON format.</param>
    /// <param name="sourceName">the exact source name or "select" to prompt dialog selection; "default" to use default source; "current" refers to current opened source if any.</param>
    /// <param name="showUI">set to true to use scanner UI or false to hide the UI. Set to true for maximum compatibility.</param>
    /// <param name="modalUI">whether the scanner UI should be modal. Set to to true if you are not sure.</param>
    /// <returns>Scan result in JSON or null if user cancels.</returns>
    public String ScanAndReturnRaw(String scanRequestInJson, String sourceName, Boolean showUI, Boolean modalUI)
    {
        ensureScanFuncsCallInTheSameThread();
        String rawResult = CallFunc(FUNC_twain_scan, appId,
            // FUJITSU fi-5120Cdj waits forever if we use windowHandle (from JFrame) here
                0, // windowHandle,
                scanRequestInJson, sourceName, showUI, modalUI);
        if (rawResult != null && rawResult.Contains("failed to open data source: TWRC_CANCEL"))
        {
            return null;
        }
        return rawResult;
    }
    
    /// <summary>
    /// Retrieve list of sources (i.e., devices) optionally with caps; the default source has "default" set to true.
    /// </summary>
    /// <param name="nameOnly">if true, return source with name only otherwise return with capabilities info</param>
    /// <param name="capsToRetrieve">only effective if nameOnly is false - cap name or code separated by comma or 'all' to list all caps supported.</param>
    /// <param name="detectDeviceType">detect whether the device has ADF and/or flatbed.</param>
    /// <param name="excludeTwainDsOnWia">exclude WIA synthesized sources</param>
    /// <returns></returns>
    public List<Source> ScanListSources(bool nameOnly, String capsToRetrieve, bool detectDeviceType, bool excludeTwainDsOnWia)
    {
        String raw = CallFunc(FUNC_twain_list_sources, appId, windowHandle, nameOnly, capsToRetrieve, false, detectDeviceType, excludeTwainDsOnWia);
        if (raw == null || raw.StartsWith("<error"))
        {
            throw new ImagingException("Failed to list sources: " + raw);
        }

        List<Source> sources = new List<Source> { };
        if (nameOnly)
        {
            if (raw.Trim().Length == 0)
            {
                return sources;
            }

            string[] sourceNames = raw.Trim().Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            foreach (string sourceName in sourceNames)
            {
                sources.Add(new Source(sourceName));
            }
        }
        else
        {
            // Console.WriteLine(raw);
            sources = Source.createSourceList(ScanUtils.parseJson(raw));
        }

        return sources;
    }

    /// <summary>
    /// Retrieve list of sources (i.e., devices) with all capabilities; the default source has "default" set to true.
    /// </summary>
    /// <returns>Sources with capabilities</returns>
    public List<Source> ScanListSourcesWithFullDetails()
    {
        return ScanListSources(false, "all", true, false);
    }

    /// <summary>
    /// Lists all sources with names only.
    /// </summary>
    /// <returns>Sources with names only.</returns>
    public List<Source> ScanListSources()
    {
        return ScanListSources(true, null, false, false);
    }

    /// <summary>
    /// Gets the name of the default source or null if none presents.
    /// </summary>
    /// <returns>the name of the default source or null if none presents.</returns>
    public String scanGetDefaultSourceName()
    {
        ensureScanFuncsCallInTheSameThread();
        String result = CallFunc(FUNC_twain_get_default_source_name, appId, windowHandle);
        return result;
    }

    /// <summary>
    /// Prompts system device selection dialog for the user to select the default device.
    /// </summary>
    /// <param name="defaultSelectedSourceName">the name of the source to be pre-selected or null to use the system default.</param>
    /// <returns>the name of the source selected or empty string if user cancels or error occurs.</returns>
    public String ScanSelectDefaultSource(String defaultSelectedSourceName = null)
    {
        ensureScanFuncsCallInTheSameThread();
        String result = CallFunc(FUNC_twain_select_default_source, appId, windowHandle, defaultSelectedSourceName);
        return result;
    }

    /// <summary>
    /// Retrieve the source info optionally with caps.
    /// </summary>
    /// <param name="sourceName">the exact source name or "default" to open default source; "select" to prompt dialog selection</param>
    /// <param name="nameOnly">if true, return the device name only</param>
    /// <param name="capsToRetrieve">only effective if nameOnly is false - can be cap name or code separated by comma or 'all' to list all caps supported.</param>
    /// <param name="getCurrent">true to return current value false to return current value as well as value range if available.</param>
    /// <param name="detectDeviceType">detect whether the device has ADF and/or flatbed.</param>
    /// <param name="capsToSet">optionally, you may set capabilities before getting capabilities, e.g, 'ICAP_PIXELTYPE: TWPT_GRAY,TWPT_BW; ICAP_XSCALING/RESET; ICAP_XRESOLUTION: 200'</param>
    /// <returns>The source object</returns>
    public Source GetSource(String sourceName, Boolean nameOnly, String capsToRetrieve = null,
                        Boolean getCurrent = false, Boolean detectDeviceType = false, String capsToSet = null)
    {
        ensureScanFuncsCallInTheSameThread();
        String raw = CallFunc(FUNC_twain_get_source, appId, windowHandle, sourceName, nameOnly,
                capsToRetrieve, getCurrent, detectDeviceType, capsToSet);
        if (raw == null || raw.StartsWith("<error"))
        {
            throw new ImagingException("Failed to get source: " + raw);
        }

        XmlReader reader = JsonReaderWriterFactory.CreateJsonReader(Encoding.Default.GetBytes(raw), new XmlDictionaryReaderQuotas());
        return new Source(XElement.Load(reader));
    }


    /// <summary>
    /// Invokes a function
    /// </summary>
    /// <param name="functionName"></param>
    /// <param name="args"></param>
    /// <returns></returns>
    public static String CallFunc(String functionName, params Object[] args) {
        String jsonRequest = null;
        using(MemoryStream ms = new MemoryStream()) {
            XmlDictionaryWriter writer = JsonReaderWriterFactory.CreateJsonWriter(ms);
            writer.WriteStartDocument(true);
            writer.WriteStartElement("root");
            writer.WriteAttributeString("type", "object");
            writer.WriteElementString("function", functionName);
            writer.WriteStartElement("args");
            writer.WriteAttributeString("type", "array");
            for(int i = 0; i < args.Length; i++) {
                Object arg = args[i];
                writer.WriteStartElement("item");
                if (arg == null)
                {
                    writer.WriteAttributeString("type", "null");
                    // writer.WriteString(""); // Don't write for null
                }
                else
                {
                    TypeCode tc = Type.GetTypeCode(arg.GetType());
                    if (tc == TypeCode.Int16 || tc == TypeCode.Int32 || tc == TypeCode.Int64)
                    {
                        writer.WriteAttributeString("type", "number");
                        writer.WriteString(arg.ToString());
                    }
                    else if (tc == TypeCode.Boolean)
                    {
                        writer.WriteAttributeString("type", "boolean");
                        writer.WriteString( (Boolean)arg ? "true" : "false");
                    }
                    else
                    {
                        writer.WriteString(arg.ToString());
                    }
                        
                }
                writer.WriteEndElement();
            }
            writer.WriteEndElement(); // args array
            writer.WriteEndElement(); // root
            writer.WriteEndDocument();
            writer.Flush();

            ms.Flush();
            ms.Position = 0;
            jsonRequest = new StreamReader(ms).ReadToEnd();
        }
            
        if(jsonRequest == null || jsonRequest.Length == 0) {
            throw new ArgumentException("Invalid requst: " + jsonRequest);
        }

        IntPtr result = IntPtr.Zero;
        try
        {
            result = Is64BitProcess ?
                ImgDll64.asprise_scan_call_func(jsonRequest) :
                ImgDll32.asprise_scan_call_func(jsonRequest);
        }
        catch (DllNotFoundException dllNotFound)
        {
            MessageBox.Show("Please copy " + getImgDllName() + " to any folder in PATH. For example, any of the following: \n" 
                + Environment.GetEnvironmentVariable("SystemRoot") + "\\system32\n"
                + Directory.GetCurrentDirectory(), "Dll not found in PATH", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            throw dllNotFound;
        }

        if (result == IntPtr.Zero)
        {
            return null;
        }
        String s = Marshal.PtrToStringAnsi(result);
        deleteC(result);
        return s;
    }

    /// <summary>
    /// Configure logging settings.
    /// </summary>
    /// <param name="level">Any of LOG_LEVEL_INFO, etc.</param>
    /// <param name="logFilePath">path to the target output file or special values: "stdout", "stderr" for console logging, null or empty string to disable logging.</param>
    public static void configureNativeLogging(int level, String logFilePath)
    {
        if (Is64BitProcess)
        {
            ImgDll64.asprise_scan_configure_logging(level, logFilePath);
        }
        else
        {
            ImgDll32.asprise_scan_configure_logging(level, logFilePath);
        }
    }

    /// <summary>
    /// The library version.
    /// </summary>
    /// <returns>The library version.</returns>
    public static string GetLibraryVersion()
    {
        return CallFunc(FUNC_version);
    }

    /// <summary>
    /// Build info such as date.
    /// </summary>
    /// <returns></returns>
    public static String GetLibraryBuildInfo()
    {
        return CallFunc(FUNC_buildInfo);
    }

    /// <summary>Input the license code </summary>
    /// <param name="licenseeName">Licensee name</param>
    /// <param name="licenseCode">License code</param>
    public static void InputLicense(string licenseeName, string licenseCode)
    {
        if (Is64BitProcess)
        {
            ImgDll64.asprise_scan_input_license(licenseeName, licenseCode);
        }
        else
        {
            ImgDll32.asprise_scan_input_license(licenseeName, licenseCode);
        }
    }

    /// <summary>Finds the asprise_imaging dll in system path or from bundle and return the path to the dll. </summary>
    public static string getDll()
    {
        // 1. Search path.
        string dllFilePath = AspriseImaging.getImgDllFromPath();
        if (dllFilePath == null)
        {
            // 2. Then parent folders
            string parentFolder = AspriseImaging.detectImgDllInParentFolders();
            if (parentFolder != null)
            {
                AspriseImaging.addToSystemPath(parentFolder);
                // log("Folder containing ocr dll detected: " + parentFolder);
            }
        }
        dllFilePath = AspriseImaging.getImgDllFromPath();
        if (dllFilePath != null)
        {
            return dllFilePath;
        }

        return getImgDllFromPath();
    }

    /// <summary>
    /// Search PATH and return the location of the ocr dll.
    /// </summary>
    /// <returns></returns>
    protected static string getImgDllFromPath()
    {
        return searchFileInPath(getImgDllName());
    }

    /// <summary>
    /// The simple name of the imaging dll file.
    /// </summary>
    /// <returns></returns>
    public static string getImgDllName()
    {
        return Is64BitProcess ? DLL_NAME_64 : DLL_NAME_32;
    }

    /// <summary>
    /// Search the ancester directories and return the directory that contains imaging dll or null if not found.
    /// </summary>
    /// <returns></returns>
    private static string detectImgDllInParentFolders()
    {
        string folder = AppDomain.CurrentDomain.BaseDirectory;
        while (true)
        {
            if (File.Exists(Path.Combine(folder, getImgDllName())))
            {
                return folder;
            }
            else
            {
                folder = Path.GetDirectoryName(folder);
                if (folder == null)
                {
                    break;
                }
            }
        }
        return null;
    }

    /// <summary>
    /// Returns the absolute path of the first occurrence
    /// </summary>
    /// <param name="fileSimpleName"></param>
    /// <returns></returns>
    protected static string searchFileInPath(string fileSimpleName)
    {

        string path = getSystemPath();
        string[] folders = path.Split(new string[] { ";" }, StringSplitOptions.RemoveEmptyEntries);
        // insert current dir to folders
        string[] extended = new string[folders.Length + 1];
        extended[0] = Directory.GetCurrentDirectory();
        Array.Copy(folders, 0, extended, 1, folders.Length);
        folders = extended;
        for (int i = 0; i < folders.Length; i++)
        {
            string folder = folders[i];
            folder = folder.Replace('/', '\\');
            if (!folder.EndsWith("\\"))
            {
                folder += "\\";
            }
            string file = folder + fileSimpleName;
            if (File.Exists(file))
            {
                return file;
            }
        }
        return null;
    }

        /// <summary>
        /// Running in 64bit mode?
        /// </summary>
    public static bool Is64BitProcess
    {
        get { return IntPtr.Size == 8; }
    }

    /// <summary>
    /// Performs native C/C++ delete
    /// </summary>
    /// <param name="ptr">pointer</param>
    protected static void deleteC(IntPtr ptr)
    {
        if (Is64BitProcess)
        {
            ImgDll64.asprise_scan_util_delete_chars(ptr.ToInt64());
        }
        else
        {
            ImgDll32.asprise_scan_util_delete_chars(ptr.ToInt64());
        }
    }

        /// <summary>
        /// Returns the system path
        /// </summary>
        /// <returns>System path</returns>
    protected static string getSystemPath()
    {
        return Environment.GetEnvironmentVariable("PATH");
    }

    /// <summary>
    /// Adds the given directory to the PATH variable.
    /// </summary>
    /// <param name="dir">The folder to be added to PATH</param>
    public static void addToSystemPath(string dir)
    {
        Environment.SetEnvironmentVariable("PATH", Environment.GetEnvironmentVariable("PATH") + ";" + dir);
    }

    /// <summary>
    /// Returns the ToString() for non-null object or ""/"null" for null.
    /// </summary>
    /// <param name="obj">target object</param>
    /// <param name="nullAsEmpty">true to return "" for null; false "null"</param>
    /// <returns></returns>
    protected static string objectToString(object obj, bool nullAsEmpty = true)
    {
        if (obj == null)
        {
            return nullAsEmpty ? "" : "null";
        }
        return obj.ToString();
    }
    
    /// <summary>
    /// Returns the first non-null object or null if all arguments are null.
    /// </summary>
    /// <param name="o"></param>
    /// <param name="others"></param>
    /// <returns></returns>
    protected static object firstNonNull(object o, params object[] others)
    {
        if (o != null)
        {
            return o;
        }

        for (var i = 0; others != null && i < others.Length; i++)
        {
            if (others[i] != null)
            {
                return others[i];
            }
        }

        return null;
    }

        /// <summary>
        /// Converts utf8 encoded string to unicode
        /// </summary>
        /// <param name="utf8String"></param>
        /// <returns></returns>
    protected static string utf8ToUnicode(string utf8String)
{
    Encoding ansiEncoding = Encoding.GetEncoding(1252);

    byte[] utf8Bytes = new byte[utf8String.Length];
    for (int i = 0; i < utf8String.Length; ++i)
    {
        utf8Bytes[i] = ansiEncoding.GetBytes(utf8String.Substring(i, 1))[0];
    }

    return Encoding.UTF8.GetString(utf8Bytes, 0, utf8Bytes.Length);
}

/// <summary>return true if the given string is null or of lenght 0. </summary>
    public static bool isEmpty(string s)
        {
            return s == null || s.Trim().Length == 0;
        }
    }
}
